import os
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from sklearn.model_selection import train_test_split

from ceruleo.dataset.catalog.CMAPSS import CMAPSSDataset, sensor_indices
from ceruleo.dataset.catalog.PHMDataset2018 import FailureType, PHMDataset2018
from ceruleo.dataset.transformed import TransformedSerializedDataset
from ceruleo.dataset.ts_dataset import AbstractPDMDataset, FoldedDataset
from ceruleo.transformation import Transformer
from ceruleo.transformation.features.scalers import MinMaxScaler
from ceruleo.transformation.features.selection import ByNameFeatureSelector

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


class MockDataset(AbstractPDMDataset):
    def __init__(self, nlives: int):
        super().__init__()
        self.lives = [
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, (i + 1) * 100, 50),
                    "feature2": np.linspace(-25, (i + 1) * 500, 50),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
            for i in range(nlives - 1)
        ]

        self.lives.append(
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, 5 * 100, 50),
                    "feature2": np.linspace(-25, 5 * 500, 50),
                    "feature3": np.linspace(-25, 5 * 500, 50),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
        )

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class TestDataset:
    def test_dataset(self):
        ds = MockDataset(5)
        columns = [set(["feature1", "feature2", "RUL"]) for i in range(4)]
        columns.append(set(["feature1", "feature2", "feature3", "RUL"]))
        for life, columns in zip(ds, columns):
            assert set(life.columns) == columns
        assert ds.n_time_series == 5

        p = ds.to_pandas()
        assert p.shape[0] == 50 * 5
        assert set(ds.common_features()) == set(["feature1", "feature2", "RUL"])

        folded = ds[[3, 2, 1]]
        assert isinstance(folded, FoldedDataset)
        assert folded[0][["feature1", "feature2"]].equals(
            ds[3][["feature1", "feature2"]]
        )
        assert not folded[1][["feature1", "feature2"]].equals(
            ds[3][["feature1", "feature2"]]
        )

        two_folds = ds[[3, 2, 1]][[1, 0]]
        assert two_folds.original_indices() == [2, 3]

    def test_CMAPSSDataset(self):
        ds = CMAPSSDataset()
        assert len(ds) == 709
        life = ds[0]
        assert len(life.columns) == 28

    def test_transformed_dataset(self):
        dataset = MockDataset(nlives=5)

        pipe = ByNameFeatureSelector(features=["feature1"])
        pipe = MinMaxScaler(range=(-1, 1))(pipe)

        target_pipe = ByNameFeatureSelector(features=["RUL"])

        transformer = Transformer(pipelineX=pipe, pipelineY=target_pipe)

        transformer.fit(dataset)

        transformed_dataset = dataset.map(transformer)
        transformed_dataset.preload()

        X = transformed_dataset.get_features_of_life(0, pandas=True)
        assert isinstance(X, pd.DataFrame)

        X = transformed_dataset.get_features_of_life(0, pandas=False)
        assert isinstance(X, np.ndarray)

        path = Path("./saved_dataset").resolve()
        transformed_dataset.save(path)

        transformed_serialized_dataset = TransformedSerializedDataset(path)
        assert len(transformed_serialized_dataset) == len(transformed_dataset)

        assert np.all(transformed_serialized_dataset[0][0] == transformed_dataset[0][0])
        assert np.all(transformed_serialized_dataset[1][0] == transformed_dataset[1][0])

        dataset = MockDataset(nlives=30)
        train_ds, test_ds = train_test_split(dataset, train_size=0.8)
        train_ds, val_ds = train_test_split(train_ds, train_size=0.8)

        len_train_ds = len(train_ds)
        len_val_ds = len(val_ds)
        len_test_ds = len(test_ds)

        pipe = ByNameFeatureSelector(features=["feature1"])
        pipe = MinMaxScaler(range=(-1, 1))(pipe)

        target_pipe = ByNameFeatureSelector(features=["RUL"])

        transformer = Transformer(pipelineX=pipe, pipelineY=target_pipe)

        transformer.fit(train_ds)

        train_path = Path("./saved_dataset/train/").resolve()
        val_path = Path("./saved_dataset/val/").resolve()
        test_path = Path("./saved_dataset/test/").resolve()
        train_ds.map(transformer).save(train_path)
        val_ds.map(transformer).save(val_path)
        test_ds.map(transformer).save(test_path)

        transformed_serialized_dataset = TransformedSerializedDataset(train_path)
        n = []
        for X, y, sw in transformed_serialized_dataset:
            n.append(y.shape[0])

        assert len(n) == len_train_ds
        assert len(transformed_serialized_dataset) == len_train_ds

        transformed_serialized_dataset = TransformedSerializedDataset(val_path)
        assert len(transformed_serialized_dataset) == len_val_ds

        transformed_serialized_dataset = TransformedSerializedDataset(test_path)
        assert len(transformed_serialized_dataset) == len_test_ds


class TestAnalysis:
    def test_analysis(self):
        class MockCorruptedDataset(AbstractPDMDataset):
            def __init__(self):
                super().__init__()
                self.lives = [
                    pd.DataFrame(
                        {
                            "feature1": [np.nan, np.nan, 3, 4],
                            "feature2": [np.nan, 2, 3, 4],
                            "RUL": [1, 2, 3, 4],
                        }
                    ),
                    pd.DataFrame(
                        {
                            "feature1": [0, np.nan, 3, 4],
                            "feature2": [2, 2, 2, 2],
                            "RUL": [1, 2, 3, 4],
                        }
                    ),
                ]

            def get_time_series(self, i: int):
                return self.lives[i]

            @property
            def rul_column(self):
                return "RUL"

            @property
            def n_time_series(self):
                return len(self.lives)

        # ds = MockCorruptedDataset()
        # df, null_per_life = null_proportion(ds)
        # assert null_per_life['feature1'][0] == 0.5
        # assert null_per_life['feature2'][1] == 0

        # df, var_per_life = variance_information(ds)
        # assert var_per_life['feature2'][1] == 0


class TestCMAPSS:
    def test_CMAPSS(self):
        train_dataset = CMAPSSDataset(train=True, models=["FD001"])
        sensors_from_article = [2, 3, 4, 7, 8, 9, 11, 12, 13, 14, 15, 17, 20, 21]

        assert len(train_dataset) == 100

        features = np.array([train_dataset[0].columns[i] for i in sensor_indices])

        labels_true = np.array([f"SensorMeasure{f}" for f in sensors_from_article])

        assert (features == labels_true).all()


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test should not run in Github Actions.")
def test_PHMDataset2018():
    data = PHMDataset2018()
    assert isinstance(data, PHMDataset2018)
    assert len(data) > 1
    assert isinstance(data[0], pd.DataFrame)

    data1 = PHMDataset2018(tools=["01M01", "04M01"])
    assert len(data1) < len(data)

    data2 = PHMDataset2018(
        failure_types=FailureType.FlowCoolPressureDroppedBelowLimit,
        tools=["01M01", "04M01"],
    )
    assert len(data2) < len(data1)



from typing import Tuple
import tensorflow as tf
from ceruleo.models.keras.layers import ExpandDimension, ResidualShrinkageBlock
from keras import Input, Model, Sequential
from keras.layers import (

    BatchNormalization,

    Dense,

    Conv1D,
    ReLU,
    LSTM,
    Bidirectional,
    Concatenate,
    GlobalAveragePooling1D, 
    MaxPooling1D

)
from typing import Tuple


def MSWRLRCN(input_shape: Tuple[int, int]):
    """MSWR-LRCN: A new deep learning approach to remaining useful life estimation of bearings
        Yongyi Chen, Dan Zhang, Wen-an Zhang
    
    [Reference](https://doi.org/10.1016/j.conengprac.2021.104969)
    """

    def ConvBlock(n_filters: int, kernel_size: int):
        return Sequential(
            [
                Conv1D(n_filters, kernel_size, padding='same'),
                BatchNormalization(),
                ReLU(),
                MaxPooling1D(2),
            ]
        )

    input = Input(input_shape)
    x = input
    x1 = ConvBlock(input_shape[-1], 16)(x)
    x = ConvBlock(input_shape[-1], 32)(x1)
    x = ResidualShrinkageBlock()(x)
    x = GlobalAveragePooling1D()(x)
    
    x1 = Bidirectional(LSTM(20))(x1)
    x = Concatenate()([x1, x])
    x = ExpandDimension(2)(x)
    
    x = Bidirectional(LSTM(20, return_sequences=True))(x)
    x = tf.keras.activations.softsign(x)
    x = Bidirectional(LSTM(20, return_sequences=True))(x)
    x = tf.keras.activations.softsign(x)
    x = Bidirectional(LSTM(20))(x)
    x = tf.keras.activations.softsign(x)
    x = Dense(64, activation='relu' )(x)
    x = Dense(1, activation='relu' )(x)
    return Model(inputs=[input], outputs=[x])




from typing import Optional

import numpy as np
from ceruleo.iterators.batcher import get_batcher
from ceruleo.models.model import TrainableModel
from torchsummary import summary as model_summary
from tqdm.auto import tqdm

import torch
import torch.nn.functional as F

LOSSES = {
    'mae': F.l1_loss,
    'mse': F.mse_loss
}


class TorchTrainableModel(TrainableModel):
    def __init__(self, learning_rate: float = 0.001, loss: str = 'mse', **kwargs):
        super(TorchTrainableModel, self).__init__(**kwargs)
        self.learning_rate = learning_rate
        self._optimizer = None
        self._scheduler = None
        self.loss = LOSSES[loss]

    def build_optimizer(self):
        raise NotImplementedError

    def build_scheduler(self):
        raise None

    def _create_batchers(self, train_dataset, validation_dataset):
        train_batcher, val_batcher = super(TorchTrainableModel, self)._create_batchers(
            train_dataset, validation_dataset)
        train_batcher.restart_at_end = False
        return train_batcher, val_batcher

    @property
    def optimizer(self):
        if self._optimizer is None:
            self._optimizer = self.build_optimizer()
        return self._optimizer

    @property
    def scheduler(self):
        if self._scheduler is None:
            self._scheduler = self.build_scheduler()
        return self._scheduler

    def summary(self):
        model_summary(self.model, self.input_shape)

    def predict(self, dataset, step=None, batch_size=512, evenly_spaced_points: Optional[int] = None):
        step = self.computed_step if step is None else step

        batcher = get_batcher(dataset,
                              self.window,
                              batch_size,
                              self.transformer,
                              step,
                              shuffle=False,
                              output_size=self.output_size,
                              cache_size=self.cache_size,
                              evenly_spaced_points=evenly_spaced_points,
                              restart_at_end=False)

        y_pred = []
        for X, _ in batcher:
            y_pred.append(self.model(torch.Tensor(X)).detach().numpy())

        return np.concatenate(y_pred)

    def fit(self, train_dataset, validation_dataset, epochs: int = 100, refit_transformer: bool = True,
            print_summary: bool = True):
        if refit_transformer:
            self.transformer.fit(train_dataset)

        if print_summary:
            self.summary()
        train_batcher, val_batcher = self._create_batchers(
            train_dataset, validation_dataset)
        for i in tqdm(range(epochs)):
            pbar = tqdm(total=len(train_batcher))
            train_loss = []
            for X, y in train_batcher:
                pbar.set_description('epoch %i' % (i+1))
                self.optimizer.zero_grad()
                y_pred = self.model(torch.Tensor(X))

                single_loss = self.loss(y_pred, torch.Tensor(y))
                single_loss.backward()
                self.optimizer.step()
                train_loss.append(single_loss.detach().numpy())
                pbar.set_postfix(train_loss=np.mean(train_loss))
                pbar.update()

            val_loss = []
            for X, y in val_batcher:
                y_pred = self.model(torch.Tensor(X))
                single_loss = self.loss(y_pred, torch.Tensor(y))
                val_loss.append(single_loss.detach().numpy())

            pbar.set_postfix(train_loss=np.mean(train_loss),
                             val_loss=np.mean(val_loss))

            pbar.close()



from copy import copy
from typing import Callable, List, Optional, Tuple, Union

import matplotlib
import matplotlib.patheffects as PathEffects
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from datetime import timedelta
from typing import Iterable 
import pandas as pd

def add_vertical_line(ax, v_x, label, color, line, n_lines):

    miny, maxy = ax.get_ylim()
    ax.axvline(v_x, label=label, color=color)
    txt = ax.text(
        v_x,
        miny + (maxy - miny) * (0.5 + 0.5 * (line / n_lines)),
        label,
        color=color,
        fontweight="semibold",
    )
    txt.set_path_effects([PathEffects.withStroke(linewidth=2, foreground="w")])


def durations_histogram(
    datasets: Union[AbstractPDMDataset, List[AbstractPDMDataset]],
    *,
    label: Union[str, List[str]],
    xlabel: str = 'Cycle Duration',    
    bins: int = 15,
    units: str = "m",
    vlines: List[Tuple[float, str]] = [],
    ax:Optional[matplotlib.axes.Axes]=None,
    add_mean: bool = True,
    add_median: bool = True,
    transform: Callable[[float], float] = lambda x: x,
    threshold: float = np.inf,
    color=None,
    **kwargs,
) ->  matplotlib.axes.Axes:
    """Generate an histogram from the lives durations of the dataset

    Example:
    '''
        durations_histogram(
            [train_dataset,validation_dataset],
            label=['Train','Validation'],
            xlabel='Unit Cycles',
            units='cycles',
            figsize=(17, 5));
    '''

    Parameters:
        datasets: Dataset from which take the lives durations
        xlabel: Label of the x axis, by default Cycle Duration
        label: Label of each dataset to use as label in the boxplot, by default 1
        bins:  Number of bins to compute in the histogram, by default 15
        units: Units of time of the lives. Useful to generate labels, by default m
        vlines: Vertical lines to add to the figure in the form [(x_coordinate, label)]
        ax: Axis where to draw the plot. If missing a new figure will be created
        add_mean: Whether to add a vertical line with the mean value, by default True
        add_median: whether to add a vertical line with the median value, by default True
        transform: A function to transform each duration, by default identity transform
        threshold: Includes duration less than the threshold, by default np.inf

    Returns:
        The axis in which the histogram was created

    """
    if isinstance(datasets, list):
        assert isinstance(label,list)
        assert len(datasets) == len(label)
        label_list = label
    else:
        assert isinstance(label, str)
        datasets = [datasets]
        label_list = [label]

    durations = []
    for ds in datasets:
        durations.append([transform(duration) for duration in ds.durations()])

    return histogram_from_durations(
        durations,
        xlabel=xlabel,
        label=label_list,
        bins=bins,
        units=units,
        vlines=vlines,
        ax=ax,
        add_mean=add_mean,
        add_median=add_median,
        threshold=threshold,
        color=color,
        **kwargs,
    )



def histogram_from_durations(
    durations: List[List[float]],
    xlabel: str,
    label:  List[str],
    bins: int = 15,
    units: str = "m",
    vlines: List[Tuple[float, str]] = [],
    ax=None,
    add_mean: bool = True,
    add_median: bool = True,
    threshold:  float = np.inf,
    color=None,
    alpha=1.0,
    **kwargs
) ->  matplotlib.axes.Axes:
    
    if ax is None:
        _, ax = plt.subplots(1, 1, **kwargs)

    
    assert isinstance(label, list)
    assert len(durations) == len(label)
 

    elem_is_timedelta = isinstance(durations[0][0], timedelta)

    


    for l, dur in zip(label, durations):
        if len(l) > 0:
            l += " "
        vlines = copy(vlines)
        durations_array = np.array(dur)
        if elem_is_timedelta:
            durations_array = durations /  pd.Timedelta(1, units)
        if add_mean:
            vlines.append((float(np.mean(durations_array)), l + "Mean"))
        if add_median:
            vlines.append((float(np.median(durations_array)), l + "Median"))
        durations_array = durations_array[durations_array<threshold]
        ax.hist(durations_array, bins, color=color, alpha=alpha, label=l)

    ax.set_xlabel(xlabel + f" [{units}]" if units else "")
    ax.set_ylabel("Number of run-to-failure cycles")

    colors = sns.color_palette("hls", len(vlines))
    for i, (v_x, l) in enumerate(vlines):
        vertical_label = f"{l}: {v_x:.2f} [{units}]"
        add_vertical_line(ax, v_x, vertical_label, colors[i], i, len(vlines))
    ax.legend()

    return ax


def durations_boxplot(
    datasets: Union[AbstractPDMDataset, List[AbstractPDMDataset]],
    xlabel: Union[str, List[str]],
    ylabel: str = 'Cycle Duration',
    ax:Optional[matplotlib.axes.Axes]=None,
    hlines: List[Tuple[float, str]] = [],
    units: str = "m",
    transform: Callable[[float], float] = lambda x: x,
    maxy: Optional[float] = None,
    **kwargs,
) ->  matplotlib.axes.Axes:
    """Generate boxplots of the lives duration

    Example:

        ax = durations_boxplot(
            [train_dataset, validation_dataset],
            xlabel=['Train', 'Validation'],
            ylabel='Unit Cycles',
            figsize=(17, 5))

    Parameters:
        datasets: Dataset from which take the lives durations
        xlabel:  Label of each dataset to use as label in the boxplot
        ylabel: Label of the y axis
        ax: Axis where to draw the plot.If missing a new figure will be created
        hlines: Horizontal lines to add to the figure in the form [(y_coordinate, label)]
        units: Units of time of the lives. Useful to generate labels
        transform: A function to transform each duration
        maxy: Maximum y value of the plot

    Returns:
        Axis where plot has been drawn
    """
    if isinstance(datasets, list):
        assert isinstance(xlabel, list)
        assert isinstance(datasets, list)
        assert len(datasets) == len(xlabel)
        xlabel_list = xlabel
        datasets_list = datasets
    else:
        assert isinstance(xlabel, str)
        datasets_list = [datasets]
        xlabel_list = [xlabel]

    durations = []
    for ds in datasets_list:
        durations.append([transform(duration) for duration in ds.durations()])

    return boxplot_from_durations(
        durations,
        xlabel=xlabel_list,
        ylabel=ylabel,
        ax=ax,
        hlines=hlines,
        units=units,
        maxy=maxy,
        **kwargs,
    )


def boxplot_from_durations(
    durations: List[List[float]],
    xlabel:  List[str],
    ylabel: str,
    ax=None,
    hlines: List[Tuple[float, str]] = [],
    units: str = "m",
    maxy: Optional[float] = None,
    **kwargs,
)->  matplotlib.axes.Axes:
    
    assert isinstance(xlabel, list)
    assert len(durations) == len(xlabel)

    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    ax.boxplot(durations, labels=xlabel)
    ax.set_ylabel(ylabel)
    if maxy is not None:
        miny, _ = ax.get_ylim()
        ax.set_ylim(miny, maxy)
    colors = sns.color_palette("hls", len(hlines))
    for i, (pos, label) in enumerate(hlines):
        ax.axhline(pos, label=f"{label}: {pos:.2f} {units}", color=colors[i])
    _, labels = ax.get_legend_handles_labels()
    if len(labels) > 0:
        ax.legend()

    return ax






import numpy as np
import pandas as pd
from ceruleo.dataset.analysis.numerical_features import NumericalFeaturesAnalysis, analyze

from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.dataset.analysis.correlation import CorrelationAnalysis, correlation_analysis
from ceruleo.transformation.features.selection import ByNameFeatureSelector
from ceruleo.transformation.functional.transformers import Transformer
from ceruleo.dataset.analysis.sample_rate import SampleRateAnalysis, sample_rate, sample_rate_summary
from ceruleo.dataset.analysis.distribution import features_divergeces

class MockDataset(AbstractPDMDataset):
    def __init__(self, nlives: int):
        super().__init__()
        self.lives = [
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, 100, 50),
                    "feature2": np.random.randint(2, size=(50,)),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
            for i in range(nlives)
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class MockDatasetTimeDeltaIndex(AbstractPDMDataset):
    def __init__(self, nlives: int):
        super().__init__()
        self.lives = [
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, 100, 50),
                    "feature2": np.random.randint(2, size=(50,)),
                    "RUL": np.linspace(100, 0, 50),
                },
                index=pd.timedelta_range(start='0 day', periods=50, freq='s')
            )
            for i in range(nlives)
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)



class TestAnalysis:
    def test_correlation(self):
        ds = MockDataset(5)
        assert isinstance(correlation_analysis(ds), CorrelationAnalysis)

        transformer = Transformer(
            pipelineX=ByNameFeatureSelector(features=['feature1', 'feature2']), 
            pipelineY=ByNameFeatureSelector(features=['RUL'])
        )

        transformed_dataset = transformer.fit_map(ds)
        assert isinstance(correlation_analysis(transformed_dataset), CorrelationAnalysis)

        q = correlation_analysis(transformed_dataset).to_pandas()
        assert isinstance(q, pd.DataFrame)

    def test_samplerate(self):
        dataset = MockDataset(5)
        sample_rates = sample_rate(dataset)
        assert isinstance(sample_rates, np.ndarray)

        assert isinstance(sample_rate_summary(dataset), SampleRateAnalysis)

        dataset = MockDatasetTimeDeltaIndex(5)

        sample_rates = sample_rate(dataset)
        assert isinstance(sample_rates, np.ndarray)

        assert isinstance(sample_rate_summary(dataset), SampleRateAnalysis)


    def test_distribution(self):
        dataset = MockDataset(5)
        assert isinstance(features_divergeces(dataset), pd.DataFrame)

    def test_analysis(self):
        dataset = MockDataset(5)
        df = analyze(dataset)
        keys = list(df.keys())
        assert isinstance(df[keys[0]], NumericalFeaturesAnalysis)



import shelve
import shutil
import uuid
from enum import Enum
from pathlib import Path

from ceruleo import CACHE_PATH


class GraphTraversalAbstractStore:
    """Abstract Cache for the graph traversal
    """
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError

    def keys(self):
        raise NotImplementedError

    def pop(self, k):
        raise NotImplementedError


class GraphTraversalCacheShelveStore:
    """Cache all the intermediate steps in a Shelve Store

    Parameters:

        cache_path: Path where the case is stored
    """
    def __init__(self, cache_path: Path = CACHE_PATH):
        filename = "".join(str(uuid.uuid4()).split("-"))
        self.cache_path = cache_path / "GraphTraversalCache" / filename / "data"
        self.cache_path.parent.mkdir(exist_ok=True, parents=True)
        self.transformed_cache = shelve.open(str(self.cache_path))

    def __getitem__(self, k):
        return self.transformed_cache[k]

    def __setitem__(self, k, v):
        self.transformed_cache[k] = v

    def close(self):
        if self.cache_path.parent.is_dir():
            self.transformed_cache.close()
            shutil.rmtree(self.cache_path.parent)

    def reset(self):
        self.transformed_cache.close()
        self.transformed_cache = shelve.open(str(self.cache_path))

    def keys(self):
        return self.transformed_cache.keys()

    def pop(self, k):
        return self.transformed_cache.pop(k)


class GraphTraversalCacheMemory:
    """Cache all the intermediate steps in RAM
    """
    def __init__(self):
        self.store = {}

    def __getitem__(self, k):
        return self.store[k]

    def __setitem__(self, k, v):
        self.store[k] = v

    def close(self):
        self.store = {}

    def reset(self):
        self.store = {}

    def keys(self):
        return self.store.keys()

    def pop(self, k):
        return self.store.pop(k)


class CacheStoreType(Enum):
    """Cache store modes
    
    Values:

        SHELVE = 1
        MEMORY = 2    
    """
    SHELVE = 1
    MEMORY = 2



from typing import Optional, Union


import pandas as pd
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.tdigest import TDigest
import numpy as np
from scipy.signal import find_peaks
from ceruleo.iterators.iterators import (
    RelativePosition,
    RelativeToEnd,
    RelativeToStart,
)


class SliceRows(TransformerStep):
    """
    Slice portion of the run-to-failure cycle
    
    Parameters:
        initial: Initial position of the slice, default RelativeToStart(0)
        final: Final position of the slice, default RelativeToEnd(0)
    """

    def __init__(
        self,
        initial: Union[int, RelativePosition] = RelativeToStart(0),
        final: Union[int, RelativePosition] = RelativeToEnd(0),
        *args,
        **kwargs
    ):

        super().__init__(*args, **kwargs)

        self.initial = initial
        self.final = final

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Slice the run-to-failure cycle

        Parameters:
            X: Input dataframe
        
        Returns:
            A dataframe with the sliced run-to-failure cycle
        """
        if isinstance(self.initial, RelativePosition):
            initial = self.initial.get(X.shape[0])
        else:
            initial = self.initial
        if isinstance(self.final, RelativePosition):
            final = self.final.get(X.shape[0])
        else:
            final = self.final

        return X.iloc[initial:final, :].copy()



from abc import ABC, abstractmethod
from typing import Iterator

import pandas as pd
from pathlib import Path

from ceruleo.dataset.builder.output import OutputMode


class CyclesSplitter(ABC):
    @abstractmethod
    def split(self, data: pd.DataFrame) -> Iterator[pd.DataFrame]:
        raise NotImplementedError


class IncreasingFeatureCycleSplitter(CyclesSplitter):
    """
    A splitter that divides a DataFrame into cycles based on changes in the value of an increasing feature.

    When the value of the increasing feature decreases, a new cycle is considered to start.

    """

    def __init__(self, increasing_feature: str):
        """Initializes the splitter with the name of the increasing feature.

        Parameters
        ----------
        increasing_feature : str
            The name of the increasing feature used for splitting.
        """
        self.increasing_feature = increasing_feature

    def split(self, data: pd.DataFrame) -> Iterator[pd.DataFrame]:
        """Splits the input DataFrame into cycles based on changes in the increasing feature.

        Parameters
        ----------
        data : pd.DataFrame
            The input DataFrame to be split.

        Yields
        ------
        Iterator[pd.DataFrame]
            An iterator of DataFrames, each containing a cycle of the input data.
        """
        restart_points = data[data[self.increasing_feature].diff() < 0].index.tolist()
        start_idx = 0
        i = 1
        for restart_idx in restart_points:
            subset = data.iloc[start_idx:restart_idx]
            yield subset.copy()
            start_idx = restart_idx
            i += 1
        yield subset.copy()


class LifeIdCycleSplitter(CyclesSplitter):
    """A splitter that divides a DataFrame into cycles based on unique life identifiers."""

    def __init__(self, life_id_feature: str):
        """Initializes the splitter with the name of the life id feature.

        Parameters
        ----------
        life_id_feature : str
            The name of the column representing the life identifier.
        """
        self.life_id_feature = life_id_feature

    def split(self, data: pd.DataFrame) -> Iterator[pd.DataFrame]:
        """Splits the input DataFrame into cycles based on unique life identifiers.

        Parameters
        ----------
        data : pd.DataFrame
            The input DataFrame to be split.

        Yields
        ------
        Iterator[pd.DataFrame]
            An iterator of DataFrames, each containing a cycle of the input data.
        """
        for life_id in data[self.life_id_feature].unique():
            subset = data[data[self.life_id_feature] == life_id]
            yield subset.copy()


class LifeEndIndicatorCycleSplitter(CyclesSplitter):
    """A splitter that divides a DataFrame into cycles based on a life end indicator feature."""

    def __init__(self, life_end_indicator_feature: str, end_value=1):
        """

        Parameters
        ----------
        life_end_indicator_feature : str
            The name of the column representing the life end indicator.
        end_value : int, optional
            The value indicating the end of a life cycle. by default 1

        """
        self.life_end_indicator_feature = life_end_indicator_feature
        self.end_value = end_value

    def split(self, data: pd.DataFrame) -> Iterator[pd.DataFrame]:
        """Splits the input DataFrame into cycles based on a life end indicator feature.

        Parameters
        ----------
        data : pd.DataFrame
            The input DataFrame to be split.

        Yields
        ------
        Iterator[pd.DataFrame]
            An iterator of DataFrames, each containing a cycle of the input data.
        """
        start_idx = 0
        for idx in data[
            data[self.life_end_indicator_feature] == self.end_value
        ].index.tolist():
            subset = data.iloc[start_idx : idx + 1]
            yield subset.copy()
            start_idx = idx + 1
        if start_idx < data.shape[0]:
            yield data.iloc[start_idx:].copy()


class FailureDataCycleSplitter(CyclesSplitter):
    """A splitter that divides a DataFrame into cycles based on a separate DataFrame containing failure data."""

    data_time_column: str
    fault_time_column: str

    def __init__(self, data_time_column: str, fault_time_column: str):
        self.data_time_column = data_time_column
        self.fault_time_column = fault_time_column

    def split(self, data: pd.DataFrame, fault: pd.DataFrame):
        data = self.merge_data_with_faults(data, fault)
        for life_index, life_data in data.groupby("fault_number"):
            if life_data.shape[0] == 0:
                continue
            yield life_data.copy()

    def merge_data_with_faults(self, data: pd.DataFrame, fault: pd.DataFrame):
        """Merge the raw sensor data with the fault information

        Parameters:

            data_file: Path where the raw sensor data is located
            fault_data_file: Path where the fault information is located

        Returns:

            df: Dataframe indexed by time with the raw sensors and faults
                The dataframe contains also a fault_number column
        """

        fault = fault.drop_duplicates(subset=[self.fault_time_column]).copy()
        fault["fault_number"] = range(fault.shape[0])
        return pd.merge_asof(
            data,
            fault,
            left_on=self.data_time_column,
            right_on=self.fault_time_column,

            suffixes=["_data", "_fault"],
            direction="forward",
        )



from typing import Optional, Union

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.tdigest import TDigest
from ceruleo.transformation.utils import  QuantileEstimator


class RobustMinMaxScaler(TransformerStep):
    """
    Scale features using statistics that are robust to outliers.

    This Scaler scales the data according to the quantile range.
    The IQR is the range between the limits provided, by default,
    1st quartile (25th quantile) and the 3rd quartile (75th quantile).

    The quantiles are approximated using tdigest

    Parameters:
        range: Desired range of transformed data.
        clip: Set to True to clip transformed values of held-out data to provided, by default True
        lower_quantile: Lower limit of the quantile range to compute the scale, by default 0.25
        upper_quantile: Upper limit of the quantile range to compute the scale, by default 0.75
        tdigest_size: Size of the t-digest structure, by default 100
        name: Name of the step, by default None
    """

    def __init__(
        self,
        *,
        range: tuple,
        clip: bool = True,
        lower_quantile: float = 0.25,
        upper_quantile: float = 0.75,
        max_workers: int = 1,
        subsample: Optional[Union[int, float]] = None,
        name: Optional[str] = None,
        prefer_partial_fit:bool = False
    ):

        super().__init__(name=name, prefer_partial_fit=prefer_partial_fit)
        self.range = range
        self.Q1 = None
        self.Q3 = None
        self.clip = clip
        self.quantile_estimator = QuantileEstimator(
            tdigest_size=50, subsample=subsample, max_workers=max_workers
        )
        self.lower_quantile = lower_quantile
        self.upper_quantile = upper_quantile

    def _compute_quantiles(self):
        self.Q1 = self.quantile_estimator.quantile(self.lower_quantile)
        self.Q3 = self.quantile_estimator.quantile(self.upper_quantile)

        self.IQR = self.Q3 - self.Q1

        self.valid_mask = self.IQR.abs() > 0.000000000001


    def partial_fit(self, df: pd.DataFrame, y=None):
        """ 
        Compute the quantiles of the dataset
        
        Parameters:
            df: The input dataset
        """
        self.quantile_estimator.update(df)
        return self

    def fit(self, df: pd.DataFrame, y=None):
        """ 
        Compute the quantiles of the dataset
        
        Parameters:
            df: The input dataset
        """
        self.quantile_estimator.update(df)
        self._compute_quantiles()
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Scale the input dataset

        Parameters:
            X: The input dataset
        
        Returns:
            A new DataFrame with the same index as the input with the
            data scaled with respect to the quantiles of the fiited dataset
        """
        if self.Q1 is None:
            self._compute_quantiles()
            

        new_X = X.copy()
        X_std = (X.loc[:, self.valid_mask] - self.Q1[self.valid_mask]) / (
            self.IQR[self.valid_mask]
        )
        new_X.loc[:, self.valid_mask] = (
            X_std * (self.range[1] - self.range[0]) + self.range[0]
        )
        new_X.loc[:, ~self.valid_mask] = 0
        if self.clip:
            return new_X.clip(self.range[0], self.range[1])
        else:
            return new_X


class MinMaxScaler(TransformerStep):
    """
    Transform features by scaling each feature to a given range.

    This transformer scales and translates each feature individually
    such that it is in the given range on the training set.

    Parameters:
        range: Desired range of transformed data.
        clip: Set to True to clip transformed values of held-out data to provided, by default True
        fillna: Wheter to fill NaN with a value
        name: Name of the step, by default None

    """

    def __init__(
        self,
        *,
        range: tuple,
        clip: bool = True,
        fillna: Optional[float] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.range = range
        self.min = range[0]
        self.max = range[1]
        self.data_min = None
        self.data_max = None
        self.clip = clip
        self.fillna = fillna

    def partial_fit(self, df: pd.DataFrame, y=None):
        """
        Compute the dataset's bounds

        Parameters:
            df: The input dataset
        """
        partial_data_min = df.min(skipna=True)
        partial_data_max = df.max(skipna=True)
        if self.data_min is None:
            self.data_min = partial_data_min
            self.data_max = partial_data_max
        else:
            self.data_min = pd.concat([self.data_min, partial_data_min], axis=1).min(
                axis=1, skipna=True
            )
            self.data_max = pd.concat([self.data_max, partial_data_max], axis=1).max(
                axis=1, skipna=True
            )
        return self

    def fit(self, df: pd.DataFrame, y=None):
        """
        Compute the dataset's bounds

        Parameters:
            df: The input dataset
        """
        self.data_min = df.min(skipna=True)
        self.data_max = df.max(skipna=True)

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Scale the input dataset

        Parameters:
            X: The input dataset
        
        Returns:
            A new DataFrame with the same index as the input with the data scaled in the range inserted in input
        """
        try:
            divisor = self.data_max - self.data_min
            
            
            
            mask = np.abs((divisor)) > 1e-25
            X = X.astype(float)
            X.loc[:, mask] = (
                (X.loc[:, mask] - self.data_min[mask])
                / divisor[mask]
                * (self.max - self.min)
            ) + self.min
            if self.fillna is not None:
                X.loc[:, ~mask] = self.fillna
        except:
            raise
        if self.clip:
            X.clip(lower=self.min, upper=self.max, inplace=True)
        return X

    def description(self):
        data = super().description()
        return (data, {"Min": self.data_min, "Max": self.data_max})


class StandardScaler(TransformerStep):
    """
    Standardize features by removing the mean and scaling to unit variance.

    Parameters:
        name: Name of the step, by default None
    """

    def __init__(self, *, name: Optional[str] = None):
        super().__init__(name=name)
        self.std = None
        self.mean = None

    def partial_fit(self, df: pd.DataFrame, y=None):
        """
        Compute mean and std of the dataset

        Parameters:
            df: The input dataset
        """
        if df.shape[0] < 15:
            return self
        partial_data_mean = df.mean()
        partial_data_std = df.std()
        if self.mean is None:
            self.mean = partial_data_mean
            self.std = partial_data_std
        else:
            self.mean = pd.concat([self.mean, partial_data_mean], axis=1).mean(axis=1)
            self.std = pd.concat([self.std, partial_data_std], axis=1).mean(axis=1)
        return self

    def fit(self, df: pd.DataFrame, y=None):
        """
        Compute mean and std of the dataset

        Parameters:
            df: The input dataset
        """
        self.mean = df.mean()
        self.std = df.std()
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Scale the input dataset

        Parameters:
            X: The input dataset

        Returns:
            A new DataFrame with the same index as the input with the data scaled to have null mean and unit variance
        """
        return (X - self.mean) / (self.std)


class RobustStandardScaler(TransformerStep):
    """
    Scale features using statistics that are robust to outliers.

    Parameters:
        quantile_range: Desired quantile range of transformed data, by defualt (0.25,0.75)
    """

    def __init__(self, *, quantile_range: tuple=(0.25, 0.75), prefer_partial_fit:bool = False, **kwargs):
        super().__init__( **kwargs,prefer_partial_fit=prefer_partial_fit)
        self.quantile_range = quantile_range
        self.quantile_estimator = QuantileEstimator()
        self.IQR = None
        self.median = None

    def fit(self, X: pd.DataFrame, y=None):
        """
        Compute the mean of the dataset

        Parameters:
            X: the input dataset   
        """
        Q1 = X.quantile(self.quantile_range[0])
        Q3 = X.quantile(self.quantile_range[1])
        self.IQR = Q3 - Q1
        self.median = X.median()

    def partial_fit(self, X: pd.DataFrame, y=None):
        """
        Compute incrementally the mean of the dataset

        Parameters:
            X: the input dataset   
        """
        if X.shape[0] < 2:
            return self

        self.quantile_estimator.update(X)

        return self

    def _compute_quantiles(self):
        self.Q1 = self.quantile_estimator.quantile(self.quantile_range[0])
        self.Q3 = self.quantile_estimator.quantile(self.quantile_range[1])

        self.IQR = self.Q3 - self.Q1

        self.median = self.quantile_estimator.quantile(0.5)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Center the input life

        Parameters:
        X: pd.DataFrame
            The input life

        Returns:
            A new DataFrame with the same index as the input with the data centered with respect to the mean of the fiited dataset
        """
        if self.IQR is None:
            self._compute_quantiles()
        return (X - self.median) / self.IQR


class ScaleInvRUL(TransformerStep):
    """
    Scale binary columns according to the inverse of the RUL.Usually this will be used before a CumSum transformation

    Parameters:
        rul_column: Column with the RUL
    """

    def __init__(self, *,rul_column: str, name: Optional[str] = None):
        super().__init__(name=name)
        self.RUL_list_per_column = {}
        self.penalty = {}
        self.rul_column_in = rul_column
        self.rul_column = None

    def partial_fit(self, X: pd.DataFrame):
        """
        Fit the scaler

        Parameters:
            X: The input dataset
        """
        if self.rul_column is None:
            self.rul_column = self.column_name(X, self.rul_column_in)
        columns = [c for c in X.columns if c != self.rul_column]
        for c in columns:
            mask = X[X[c] > 0].index
            if len(mask) > 0:
                RUL_list = self.RUL_list_per_column.setdefault(c, [])
                RUL_list.extend(
                    (
                        1
                        + (
                            X[self.rul_column].loc[mask].values
                            / X[self.rul_column].max()
                        )
                    ).tolist()
                )

        for k in self.RUL_list_per_column.keys():

            self.penalty[k] = 1 / np.median(self.RUL_list_per_column[k])

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Scale the input dataset

        Parameters:
            X: The input dataset
        
        Returns:
            A new DataFrame with the same index as the input with the data scaled with respect to the RUL
        """
        columns = [c for c in X.columns if c != self.rul_column]
        X_new = pd.DataFrame(index=X.index)
        for c in columns:
            if c in self.penalty:
                X_new[c] = X[c] * self.penalty[c]
        return X_new


class PerCategoricalMinMaxScaler(TransformerStep):
    """
    Performs a minmax scaler partition of the data trough some categorical feature

    Usually, different execution configurations lead to different scales in the features.
    Therefore, sometimes it is useful to scale the data based on a categorical feature,
    to reflect the difference in the execution parameters.

    Parameters:
        categorical_feature: str
            The name of the categorical feature whose values are going to be used to split each time-series
        scaler: The scaler to use when scaling the data, by default MinMaxScaler
        scaler_params: Parameters used when constructing the scaler, by default {}
        name: Name of the step, by default None
    """

    def __init__(
        self,
        *,
        categorical_feature: str,
        scaler: Optional[Union[MinMaxScaler, RobustMinMaxScaler]] = MinMaxScaler,
        scaler_params: dict = {},
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.categorical_feature = categorical_feature
        self.categorical_feature_name = None
        self.scaler = scaler
        self.scaler_params = scaler_params

        self.scalers = {"default": self.scaler(**self.scaler_params)}

    def partial_fit(self, X: pd.DataFrame, y=None):
        """
        Fit the scaler

        Parameters:
            X: The input dataset
        """
        if self.categorical_feature_name is None:
            self.categorical_feature_name = self.find_feature(
                X, self.categorical_feature
            )
        for category, data in X.groupby(self.categorical_feature_name):
            data = data.drop(columns=[self.categorical_feature_name])
            if category not in self.scalers:
                self.scalers[category] = self.scaler(**self.scaler_params)
            self.scalers[category].partial_fit(data)
            self.scalers["default"].partial_fit(data)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Scale the input dataset using the appropriate scaler for each category

        Parameters:
            X: The input dataset
        
        Returns:
            A new DataFrame with the same index as the input with the data scaled with respect to the categorical feature
        """
        X_new = X.drop(columns=[self.categorical_feature_name])

        for category, data in X.groupby(self.categorical_feature_name):

            data = data.drop(columns=[self.categorical_feature_name])
            scaler = (
                self.scalers[category]
                if category in self.scalers
                else self.scalers["default"]
            )  # Use a defaultdict
            X_new.loc[data.index, :] = scaler.transform(data)
        return X_new



from enum import Enum
from typing import Dict, List, Optional, Tuple, Union

import antropy as ant
import numpy as np
from pydantic import BaseModel
from scipy.stats import spearmanr
from sklearn.feature_selection import mutual_info_regression
from tqdm.auto import tqdm

from ceruleo.dataset.transformed import TransformedDataset
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.dataset.utils import iterate_over_features_and_target
import pandas as pd


class MetricType(str, Enum):
    std = "std"
    correlation = "correlation"
    autocorrelation = "autocorrelation"
    monotonicity = "monotonicity"
    number_of_unique_elements = "number_of_unique_elements"
    mutual_information = "mutual_information"
    null = "null"
    entropy = "entropy"

    @staticmethod
    def from_str(s: str) -> "MetricType":
        return MetricType(s)


class MetricValuesSummary(BaseModel):
    mean: float
    std: float
    max: float
    min: float


class NumericalFeaturesAnalysis(BaseModel):
    feature: str
    metric: Dict[MetricType, List[float]]

    def summarize(self) -> Dict[MetricType, MetricValuesSummary]:
        out = {}
        for metric in self.metric.keys():
            mean = np.nanmean(self.metric[metric])
            std = np.nanstd(self.metric[metric])
            max_ = np.nanmax(self.metric[metric])
            min_ = np.nanmin(self.metric[metric])
            out[metric] = MetricValuesSummary(mean=mean, std=std, max=max_, min=min_)
        return out
    
    def __getitem__(self, key: str) -> MetricValuesSummary:
        return self.metric[MetricType.from_str(key)]
    
    def _repr_html_(self) -> str:
        out = "<table>"
        out += "<tr><th>Metric</th><th>Mean</th><th>Std</th><th>Max</th><th>Min</th></tr>"
        for metric, summary in self.summarize().items():
            out += f"<tr><td>{metric}</td><td>{summary.mean}</td><td>{summary.std}</td><td>{summary.max}</td><td>{summary.min}</td></tr>"
        out += "</table>"
        return out


def entropy(s: np.ndarray) -> float:
    """
    Approximate entropy

    The approximate entropy quantifies the amount of regularity and the unpredictability of fluctuations over time-series data.

    Parameters:
        s: A single feature

    Returns:
        Approximate entropy of feature s
    """
    return ant.app_entropy(s)


def correlation(s: np.ndarray, y: Optional[np.ndarray] = None) -> float:
    """
    Correlation of the feature with the target

    Parameters:
        s: A single feature
        y: The RUL target

    Returns:
        Correlation between the feature and the RUL target
    """
    N = s.shape[0]
    if not (s[0] == s).all():
        corr = spearmanr(s, np.arange(N), nan_policy="omit")
        corr = corr.correlation
    else:
        corr = np.nan
    return corr


def autocorrelation(s: np.ndarray) -> float:
    """
    Autocorrelation of a feature

    Parameters:
        s: A single feature

    Returns:
        Autocorrelation of the feature
    """
    diff = np.diff(s)
    return np.sum(diff**2) / s.shape[0]


def monotonicity(s: np.ndarray) -> float:
    """
    Monotonicity of a feature, the two extreme values are 0 if the feature is constant and 1 if it is strictly monotonic.

    Parameters:
        s: A single feature

    Returns:
        Monotonicity of the feature
    """
    N = s.shape[0]
    diff = np.diff(s)
    return 1 / (N - 1) * np.abs(np.sum(diff > 0) - np.sum(diff < 0))


def n_unique(s: np.ndarray) -> int:
    """
    Number of unique values in the array

    Parameters:
        s: A single feature

    Returns:
        Number of unique values
    """
    return len(np.unique(s))


def null(s: np.ndarray) -> float:
    """
    Null percentage for a given feature

    Parameters:
        s: A feature

    Returns:
        Null percentage
    """
    return np.mean(~np.isfinite(s)) * 100


def mutual_information(x: np.ndarray, y: np.ndarray) -> float:
    """Mutual information between a feature and the target

    [Reference](Remaining Useful Life Prediction Using Ranking Mutual Information Based Monotonic Health Indicator)

    Parameters:
        x: A single feature
        y: RUL Target

    Returns:
        Mutual information between x and y

    """
    x = x.reshape(-1, 1)
    x = np.nan_to_num(x)
    return mutual_info_regression(x, y)


metrics = {
    "std": lambda x, y: np.std(x),
   
    "autocorrelation": lambda x, y: autocorrelation(x),
    "monotonicity": lambda x, y: monotonicity(x),
    "number_of_unique_elements": lambda x, y: n_unique(x),

    "null": lambda x, y: null(x),
    "entropy": lambda x, y: entropy(x),
    "mutual_information": mutual_information,
     "correlation": lambda x, y: correlation(x, y),
}


def analyze_single_cycle(
    X: np.ndarray,
    y: np.ndarray,
    out: Dict[str, Dict[MetricType, List[float]]],
    column_names: List[str],
    what_to_compute: List[str] = [],
):
    """
    Compute the analysis for a single run-to-failure cycle



    Parameters:
        X: Input Features
        y: RUL Target
        column_names: Column names of the features
        data: Initial data
        what_to_compute: Features to compute

    Returns:
        A dictionary with the analysis of the features

    """

    if len(what_to_compute) == 0:
        what_to_compute = list(sorted(metrics.keys()))
    for column_index in range(len(column_names)):
        column_name = column_names[column_index]
        for what in what_to_compute:
            x_ts = np.squeeze(X.loc[:, column_name].values)

            m = metrics[what](x_ts, y)
            metric_type = MetricType.from_str(what)
            out[column_name][metric_type].append(m)

    return out


def merge_cycle_analysis(
    data: Dict[str, Dict[MetricType, List[float]]],
) -> Dict[str, NumericalFeaturesAnalysis]:
    out = {k: NumericalFeaturesAnalysis(feature=k, metric={}) for k in data.keys()}
    for column_name in data.keys():
        for what in data[column_name]:
            metric_type = MetricType.from_str(what)
            out[column_name].metric[metric_type] = data[column_name][what]
    return out


def analyze(
    dataset: Union[TransformedDataset, AbstractPDMDataset],
    *,
    show_progress: bool = False,
    what_to_compute: List[str] = [],
) -> NumericalFeaturesAnalysis:
    """
    Compute analysis of numerical features

    Parameters:
        dataset: A transformed dataset with features and target
        show_progress: Wether to show the progress when computing the features
        what_to_compute: Elements available to compute:

            - std
            - Correlation
            - Autocorrelation
            - Monotonicity
            - Number of unique elements
            - Mutual information
            - Null
            - Entropy


    Returns:
        NumericalFeaturesAnalysis
    """

    if len(what_to_compute) == 0:
        what_to_compute = list(sorted(metrics.keys()))
    iterator = dataset
    if show_progress:
        iterator = tqdm(iterator)

    if isinstance(dataset, TransformedDataset):
        column_names = dataset.transformer.column_names
    else:
        column_names = dataset.numeric_features()

    data_per_cycle = {
        k: {MetricType.from_str(what): [] for what in what_to_compute}
        for k in column_names
    }
    for X, y in iterate_over_features_and_target(dataset):
        y = np.squeeze(y)
        analyze_single_cycle(X, y, data_per_cycle, column_names, what_to_compute)

    return merge_cycle_analysis(data_per_cycle)


def analyze_as_dataframe(
    dataset: Union[TransformedDataset, AbstractPDMDataset],
    *,
    show_progress: bool = False,
    what_to_compute: List[str] = [],
) -> pd.DataFrame:
    """
    Compute analysis of numerical features

    Parameters:
        dataset: A transformed dataset with features and target
        show_progress: Wether to show the progress when computing the features
        what_to_compute: Elements available to compute:

            - std
            - Correlation
            - Autocorrelation
            - Monotonicity
            - Number of unique elements
            - Mutual information
            - Null
            - Entropy


    Returns:
        pd.DataFrame
    """
    rr = analyze(dataset, show_progress=show_progress, what_to_compute=what_to_compute)
        
   
    out: Dict[Tuple[str, str], List[float]] = {}

  
    for k, metrics in rr.items():
        metrics_summary = metrics.summarize()
        for metric_name, metric_values in metrics_summary.items():
            
            key_mean = (metric_name.value, "Mean value across the cycles")
            key_std = (metric_name.value, "Standard deviation across the cycles")
            key_max = (metric_name.value, "Maximum value found in a cycle")
            key_min = (metric_name.value, "Minimum value found in a cycle")
            
            if key_mean not in out:
                out[key_mean] = []
                out[key_std] = []
                out[key_max] = []
                out[key_min] = []

            out[key_mean].append(metric_values.mean)
            out[key_std].append(metric_values.std)
            out[key_max].append(metric_values.max)
            out[key_min].append(metric_values.min)

    return pd.DataFrame(out, index=rr.keys())



import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K


from tensorflow.python.keras.losses import LossFunctionWrapper


def root_mean_squared_error(y_true: np.array, y_pred: np.array) -> float:
    """Root mean squared error

    Parameters:
            y_true: True RUL values
            y_pred: Predicted RUL values
    """
    return K.sqrt(K.mean(K.square(y_pred - y_true), axis=0))


def asymmetric_loss_pm(
    y_true,
    y_pred,
    *,
    theta_l,
    alpha_l,
    gamma_l,
    theta_r,
    alpha_r,
    gamma_r,
    relative_weight: bool = True,
) -> float:
    """
    Customizable Asymmetric Loss Functions for Machine Learning-based Predictive Maintenance


    Ehrig, L., Atzberger, D., Hagedorn, B., Klimke, J., & Döllner, J. (2020, October).
    Customizable Asymmetric Loss Functions for Machine Learning-based Predictive Maintenance.
    In 2020 8th International Conference on Condition Monitoring and Diagnosis
    (CMD) (pp. 250-253). IEEE.

    [Reference](https://ieeexplore.ieee.org/document/9287246)

    Parameters:
        y_true: True RUL values
        y_pred: Predicted RUL values
        theta_l: Linear to exponential change point for overpredictions (Positive)
        alpha_l: Quadratic term parameters for overpredictions
        gamma_l: Exponential term parameters for overpredictions
        theta_r: Linear to exponential change point for underpredictions
        alpha_r: Quadratic term parameters for underpredictions
        gamma_r: Exponential term parameters for underpredictions
        relative_weight: Wether to use weigthing relative to the RUL

    Returns:
        l: the loss computed
    """

    errors = y_true - y_pred
    weight = tf.abs(errors) / (
        tf.clip_by_value(y_true, clip_value_min=0.9, clip_value_max=np.inf)
    )

    ll_exp = tf.cast(K.less(errors, -theta_l), errors.dtype)
    ll_quad = (1 - ll_exp) * tf.cast(K.less(errors, 0), errors.dtype)

    lr_exp = tf.cast(K.greater(errors, theta_r), errors.dtype)
    lr_quad = (1 - lr_exp) * tf.cast(K.greater(errors, 0), errors.dtype)
    ll_exp = ll_exp * (
        alpha_l
        * theta_l
        * (theta_l + 2 * gamma_l * (K.exp((K.abs(errors) - theta_l) / (gamma_l)) - 1))
    )
    ll_quad = ll_quad * alpha_l * K.pow(errors, 2)

    lr_exp = lr_exp * (
        alpha_r
        * theta_r
        * (theta_r + 2 * gamma_r * (K.exp((errors - theta_r) / (gamma_r)) - 1))
    )
    lr_quad = lr_quad * alpha_r * K.pow(errors, 2)

    if relative_weight:
        a = tf.reduce_mean(weight * (ll_exp + ll_quad + lr_exp + lr_quad))
    else:
        a = tf.reduce_mean(ll_exp + ll_quad + lr_exp + lr_quad)

    return a


class AsymmetricLossPM(LossFunctionWrapper):
    """
    Customizable Asymmetric Loss Functions for Machine Learning-based Predictive Maintenance


    Ehrig, L., Atzberger, D., Hagedorn, B., Klimke, J., & Döllner, J. (2020, October).
    Customizable Asymmetric Loss Functions for Machine Learning-based Predictive Maintenance.
    In 2020 8th International Conference on Condition Monitoring and Diagnosis
    (CMD) (pp. 250-253). IEEE.

    [Reference](https://ieeexplore.ieee.org/document/9287246)

    Parameters:
        theta_l: Linear to exponential change point for overpredictions
        alpha_l: Quadratic term parameters for overpredictions
        gamma_l: Exponential term parameters for overpredictions
        theta_r: Linear to exponential change point for underpredictions
        alpha_r: Quadratic term parameters for underpredictions
        gamma_r: Exponential term parameters for underpredictions
        relative_weight: Wether to use weigthing relative to the RUL
    """

    def __init__(
        self,
        *,
        theta_l: float,
        alpha_l: float,
        gamma_l: float,
        theta_r: float,
        alpha_r: float,
        gamma_r: float,
        relative_weight: bool = True,
        name="asymmetric_loss_pm",
    ):
        super().__init__(
            asymmetric_loss_pm,
            theta_l=theta_l,
            alpha_l=alpha_l,
            gamma_l=gamma_l,
            theta_r=theta_r,
            alpha_r=alpha_r,
            gamma_r=gamma_r,
            relative_weight=relative_weight,
            name=name,
        )


def relative_mae(C: float = 0.9):
    """
    MAE weighted by the relative error

    Parameters:
        C: Minimal value for the RUL

    Returns:
        The loss function
    """
    mae = tf.keras.losses.MeanAbsoluteError()

    def concrete_relative_mae(y_true, y_pred):
        errors = y_true - y_pred
        sw = tf.abs(errors) / (
            tf.clip_by_value(y_true, clip_value_min=C, clip_value_max=np.inf)
        )
        return mae(y_true, y_pred, sample_weight=sw)

    return concrete_relative_mae


def relative_mse(C: float = 0.9):
    """
    MSE weighted by the relative error

    Parameters:
        C: Minimal value for the RUL

    Returns:
        The loss function
    """
    mse = tf.keras.losses.MeanSquaredError()

    def concrete_relative_mse(y_true, y_pred):
        errors = y_true - y_pred
        sw = tf.abs(errors) / (
            tf.clip_by_value(y_true, clip_value_min=C, clip_value_max=np.inf)
        )
        return mse(y_true, y_pred, sample_weight=sw)

    return concrete_relative_mse



from typing import Tuple

from ceruleo.transformation.functional.graph_utils import edges, nodes
from ceruleo.transformation.functional.transformerstep import TransformerStep


def encode_tuple(tup: Tuple):
    def get_hash(x):
        if isinstance(x, TransformerStep):
            return hash(x)
        else:
            return hash(x)

    return ",".join(list(map(lambda x: str(get_hash(x)), tup)))


def decode_tuple(s: str):
    return s.split(",")


def plot_pipeline(pipe: "TemporisPipeline", name: str):
    """Plot the transformation pipeline

    Parameters:
    
        pipe: The pipeline
        name: Title of the graphic

    Returns:

        graphic: the diagram
        
    """
    import graphviz
    from ceruleo.transformation.functional.pipeline.pipeline import \
        Pipeline

    dot = graphviz.Digraph(name, comment="Transformation graph")

    node_name = {}
    for i, node in enumerate(nodes(pipe)):
        node_name[node] = str(i) + node.name
        dot.node(str(i) + node.name, label=str(node))

    for (e1, e2) in edges(pipe):
        dot.edge(node_name[e1], node_name[e2])

    return dot



from enum import Enum
import logging
from abc import ABC, abstractmethod
from pathlib import Path

import pandas as pd

from ceruleo.dataset.ts_dataset import PDMDataset, PDMInMemoryDataset

logger = logging.getLogger(__name__)


class OutputMode(ABC):
    medatada_columns: dict

    def __init__(self):
        self.medatada_columns = {}
        self.cycles_metadata = {}

    @abstractmethod
    def store_cycle(self, cycle_id: str, df: pd.DataFrame):
        pass

    def store(self, cycle_id: str, df: pd.DataFrame, metadata: dict = {}):
        self.store_metadata(cycle_id, df, metadata)
        self.store_cycle(cycle_id, df)

    def set_metadata_columns(self, columns: dict) -> "OutputMode":
        self.medatada_columns = columns
        return self

    @abstractmethod
    def build_dataset(self, builder: "DatasetBuilder") -> PDMDataset:
        pass

    def extract_metadata(self, cycle_id:str, df: pd.DataFrame) -> dict:
        for k, v in self.medatada_columns.items():
            if v not in df.columns:
                raise ValueError(
                    f"Metadata colum {v} not present in the dataframe {df.columns}"
                )

        m = {k: df[v].iloc[0] for k, v in self.medatada_columns.items()}
        m["Number of samples"] = df.shape[0]
        m["Duration"] = df.RUL.max()
        m["Cycle ID"] = cycle_id
        return m

    def finish(self):
        pass

    def store_metadata(
        self, cycle_id: str, df: pd.DataFrame, additional_medatada: dict
    ):
        metadata = self.extract_metadata(cycle_id, df)
        metadata.update(additional_medatada)
        self.cycles_metadata[cycle_id] = metadata


class InMemoryOutputMode(OutputMode):
    def __init__(self):
        super().__init__()
        self.cycles = {}

    def store_cycle(self, cycle_id: str, df: pd.DataFrame):
        self.cycles[cycle_id] = df

    def build_dataset(self, builder: "DatasetBuilder") -> PDMDataset:
        return PDMInMemoryDataset(list(self.cycles.values()), "RUL")


class DatasetFormat(Enum):
    CSV = "csv"
    PARQUET = "parquet"
    FEATHER = "feather"


class LocalStorageOutputMode(OutputMode):
    output_path: Path
    output_format: DatasetFormat

    def __init__(
        self, output_path: Path, output_format: DatasetFormat = DatasetFormat.CSV
    ):
        super().__init__()
        self.output_path = output_path
        self.output_format = output_format

    def extract_metadata(self, cycle_id: str, df: pd.DataFrame) -> dict:
        df = super().extract_metadata(cycle_id, df)
        df["Filename"] = self.get_output_path(cycle_id, df)
        return df

    def get_output_path(self, cycle_id: str, df: pd.DataFrame) -> Path:
        return (
            self.output_path
            / "processed"
            / "cycles"
            / f"{cycle_id}.{self.output_format.value}"
        )

    def store_cycle(self, cycle_id: str, df: pd.DataFrame):
        output_cycles_path = self.output_path / "processed" / "cycles"
        if not output_cycles_path.exists():
            logger.info(f"Creating {output_cycles_path}")
            output_cycles_path.mkdir(exist_ok=True, parents=True)
        logger.info(f"Storing cycle {cycle_id}")

        write_function = {
            DatasetFormat.CSV: df.to_csv,
            DatasetFormat.PARQUET: df.to_parquet,
            DatasetFormat.FEATHER: df.to_feather,
        }
        if self.output_format not in write_function:
            raise ValueError(f"Unsupported output format {self.output_format}")

        write_function[self.output_format](
            self.get_output_path(cycle_id, df)
        )

    def finish(self):
        output_cycles_path = self.output_path / "processed" / "cycles"
        pd.DataFrame(self.cycles_metadata).T.to_csv(output_cycles_path / "cycles.csv")

    def build_dataset(self, builder: "DatasetBuilder") -> PDMDataset:
        return PDMDataset(self.output_path)



import numpy as np
import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.iterators.batcher import Batcher
from ceruleo.iterators.iterators import WindowedDatasetIterator
from ceruleo.iterators.sample_weight import (
    ExponentialDecay,
    InverseToLengthWeighted,
    NotWeighted,
    RULInverseWeighted,
)
from ceruleo.transformation import Transformer
from ceruleo.transformation.features.scalers import MinMaxScaler
from ceruleo.transformation.features.selection import ByNameFeatureSelector


class SimpleDataset(AbstractPDMDataset):
    def __init__(self):
        self.lives = [
            pd.DataFrame(
                {"feature1": np.array(range(0, 100)), "RUL": np.array(range(0, 100))}
            )
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class MockDataset(AbstractPDMDataset):
    def __init__(self, nlives: int):
        self.lives = [
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, (i + 1) * 100, 50),
                    "feature2": np.linspace(-25, (i + 1) * 500, 50),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
            for i in range(nlives - 1)
        ]

        self.lives.append(
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, 5 * 100, 50),
                    "feature2": np.linspace(-25, 5 * 500, 50),
                    "feature3": np.linspace(-25, 5 * 500, 50),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
        )

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class TestIterators:
    def test_iterators(self):
        features = ["feature1", "feature2"]
        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(x, y)
        batch_size = 15
        window_size = 5
        ds = MockDataset(5)

        transformer.fit(ds)
        b = Batcher.new(ds.map(transformer), window_size, batch_size, 1)
        X, y, w = next(b)
        assert len(y.ravel()) == batch_size
        assert X.shape[0] == batch_size
        assert X.shape[1] == window_size
        assert X.shape[2] == 2

    def test_2(self):
        dataset = SimpleDataset()
        pipe = ByNameFeatureSelector(features=["feature1"])
        y_pipe = ByNameFeatureSelector(features=["RUL"])
        transformer_raw = Transformer(pipelineX=pipe, pipelineY=y_pipe)
        transformer_raw.fit(dataset)
        it = WindowedDatasetIterator(dataset.map(transformer_raw), 5)
        X, y, sw = next(it)
        assert np.all(X == np.array([[0, 1, 2, 3, 4]]).T)
        assert y[0][0] == 4


def build_elements():
    a = np.linspace(10, 0, 11)
    b = np.vstack((a, a)).T
    c = pd.DataFrame(a, columns=["RUL"])
    d = pd.DataFrame(b, columns=["RUL", "RUL2"])
    e = pd.Series(a)
    return [a, b, c, d, e]


class TestSampleWeight:
    def test_not_weighted(self):
        nw = NotWeighted()
        for el in build_elements():
            assert nw(el, 0, None) == 1
            assert nw(el, 5, None) == 1

    def test_rul_inverse_weighted(self):
        inverse = RULInverseWeighted()
        for el in build_elements():
            assert inverse(el, 0, None) == 1.0/(10 + 1)
            assert inverse(el, 5, None) == 1.0/(5 + 1)
            assert inverse(el, 10, None) == 1.0

    def test_InverseToLengthWeighted(self):
        inverse = InverseToLengthWeighted()
        for el in build_elements():
            assert inverse(el, 0, None) == 1/(10.0)
            assert inverse(el, 5, None) ==  1/(10.0)

    def test_ExponentialDecay(self):
        exp = ExponentialDecay(near_0_at=3)
        for el in build_elements():
            assert exp(el, 0, None) < 0.00001 
            assert exp(el, 9, None) > 0.21


from typing import List, Optional

import matplotlib
import matplotlib.pyplot as plt
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.dataset.analysis.correlation import correlation_analysis


def plot_correlation_analysis(
    dataset: AbstractPDMDataset,
    corr_threshold: float = 0,
    features: Optional[List[str]] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    **kwargs,
) -> matplotlib.axes.Axes:
    """Plot the correlated features in a dataset

    Parameters:
        dataset: The dataset
        corr_threshold: Minimum threshold to consider that the correlation is high
        features: List of features
        ax: The axis where to draw

    Returns:
        The plot axis
    """

    if features is not None:
        features = list(set(features) - set(["relative_time"]))

    df = correlation_analysis(dataset, features=features).to_pandas()
    df1 = df[(df.abs_mean_correlation > corr_threshold)]

    df1.reset_index(inplace=True)
    df1.sort_values(by="mean_correlation", ascending=True, inplace=True)
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    labels = []
    for i, (_, r) in enumerate(df1.iterrows()):
        f1 = r["feature_1"]
        f2 = r["feature_2"]
        label = f"{f1}\n{f2}"
        ax.barh(
            y=i,
            width=r["mean_correlation"],
            label=label,
            xerr=r["std_correlation"],
            color="#7878FF",
        )
        labels.append(label)

    ax.axvline(x=0.90, linestyle="--")
    ax.axvline(x=-0.90, linestyle="--")

    ax.set_yticks(list(range(len(labels))))
    ax.set_yticklabels(labels)
    xticks = ax.get_xticks()

    ax.set_xticks([-1, -0.90, -0.5, 0, 0.5, 0.90, 1])
    ax.set_xlabel("Correlation")
    return ax




import logging
from pathlib import Path

import matplotlib.pyplot as plt
import tensorflow as tf
from ceruleo.graphics.results import plot_predictions
from ceruleo.iterators.utils import true_values
from ceruleo.results.results import PredictionResult
from tensorflow.keras.callbacks import Callback

logger = logging.getLogger(__name__)


class PredictionCallback(Callback):
    """
    Generate a plot after each epoch with the predictions

    Parameters:
        model: The model used predict
        output_path: Path of the output image
        dataset: The dataset that want to be plotted
    """

    def __init__(
        self,
        output_path: Path,
        dataset: tf.data.Dataset,
        units: str = "",
        filename_suffix: str = "",
    ):
        super().__init__()
        self.output_path = output_path
        self.dataset = dataset
        self.units = units
        self.suffix = filename_suffix
        if len(filename_suffix) > 0:
            self.output_path = self.output_path.with_stem(
                filename_suffix + "_" + self.output_path.stem
            )

    def on_epoch_end(self, epoch, logs={}):
        y_pred = self.model.predict(self.dataset)
        y_true = true_values(self.dataset)
        ax = plot_predictions(
            PredictionResult("Model", y_true, y_pred),
            figsize=(17, 5),
            units=self.units,
        )
        ax.legend()

        ax.figure.savefig(self.output_path, dpi=ax.figure.dpi)

        plt.close(ax.figure)



from typing import Callable, Optional
from temporis.transformation.functional.transformerstep import TransformerStep


class TransformerLambda(TransformerStep):
    def __init__(self, function:Callable,  name:Optional[str]=None):
        super().__init__(name=name)
        self.function = function

    def transform(self, X, y=None):
        return self.function(X)



from ceruleo.transformation.functional.transformers import Transformer
from ceruleo.transformation.functional.transformerstep import TransformerStep

from ceruleo.transformation.functional.concatenate import Concatenate
from ceruleo.transformation.functional.pipeline.pipeline import Pipeline


from typing import Optional


import pandas as pd
from ceruleo.transformation import TransformerStep
from ceruleo.transformation.features.tdigest import TDigest
import numpy as np
from scipy.signal import find_peaks


class MeanCentering(TransformerStep):
    """Center the data with respect to the mean"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.N = 0
        self.sum = None

    def fit(self, X: pd.DataFrame, y=None):
        """
        Compute the mean of the dataset

        Parameters:
            X: The input dataset
        """
        self.mean = X.mean()
        return self

    def partial_fit(self, X: pd.DataFrame, y=None):
        """
        Compute incrementally the mean of the dataset

        Parameters:
            X: The input life
        """
        if self.sum is None:
            self.sum = X.sum()
        else:
            self.sum += X.sum()

        self.N += X.shape[0]
        self.mean = self.sum / self.N
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Center the input life

        Parameters:
            X: The input life
            
        Returns:
            A new DataFrame with the same index as the input with the data centered with respect to the mean of the fiited dataset
        """
        return X - self.mean


class MedianCentering(TransformerStep):
    """Center the data with respect to the median"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tdigest_dict = None
        self.median = None

    def fit(self, X: pd.DataFrame, y=None):
        """
        Compute the median of the dataset

        Parameters:
            X: The input dataset
        """
        self.median = X.median()
        return self

    def partial_fit(self, X: pd.DataFrame, y=None):
        """
        Compute incrementally the median of the dataset

        Parameters:
            X: The input life
        """
        if X.shape[0] < 2:
            return self

        if self.tdigest_dict is None:
            self.tdigest_dict = {c: TDigest(100) for c in X.columns}
        for c in X.columns:
            self.tdigest_dict[c] = self.tdigest_dict[c].merge_unsorted(X[c].values)

        self.median = pd.Series(
            {
                c: self.tdigest_dict[c].estimate_quantile(0.5)
                for c in self.tdigest_dict.keys()
            }
        )
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Center the input life

        Parameters:
            X: The input life
            
        Returns:
            A new DataFrame with the same index as the input with the data centered with respect to the mean of the fiited dataset
        """
        return X - self.median


class Square(TransformerStep):
    """Compute the square of the values of each feature"""

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life with the square of the values

        Parameters:
            X: The input life

        Returns:
            A new dataframe with the same index as the input with the square of the values
        """
        return X.pow(2)


class Sqrt(TransformerStep):
    """Compute the sqrt of the values of each feature"""

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life with the sqrt of the values

        Parameters:
            X: The input life
            
        Returns:
            A new dataframe with the same index as the input with the sqrt of the values
        """
        return X.pow(1.0 / 2)


class Scale(TransformerStep):
    """
    Scale each feature by a given vaulue

    Parameters:
        scale_factor: Scale factor to apply to each feature
        name: Name of the step, by default None
    """

    def __init__(self, *, scale_factor: float, name: Optional[str] = None):
        super().__init__(name=name)
        self.scale_factor = scale_factor

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Return the scaled life

        Parameters:
            X: The input life
            
        Returns:
            Return a new DataFrame with the same index as the input with the scaled features
        """
        return X * self.scale_factor


class ExpandingCentering(TransformerStep):
    """
    Center the life using an expanding window
    """

    #.. raw:: html
    #<p>Formula: \(X - X.expanding().mean())</p>

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the live centering it using an expanding window

        Parameters:
            X: The input life
            
        Returns:
            Return a new DataFrame with the same index as the input with the data centered
        """
        return X - X.expanding().mean()


class RollingCentering(TransformerStep):
    """
    Center the life using an rolling window

    """



    def __init__(self, window: int, min_points: int, name: Optional[str] = None):
        super().__init__(name=name)
        self.window = window
        self.min_points = min_points

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the live centering it using a rolling window

        Parameters:
            X: The input life
            
        Returns:
            Return a new DataFrame with the same index as the input with the data centered
        """
        return X - X.rolling(window=self.window, min_periods=self.min_points).mean()


class ExpandingNormalization(TransformerStep):
    """Normalize the life features using an expanding window
    """

    #.. highlight:: python
    #.. code-block:: python

    #(X - X.expanding().mean()) / (X.expanding().std())


    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the live normalized it using an expanding window

        Parameters:
            X: The input life
            
        Returns:
            Return a new DataFrame with the same index as the input with the data normalized
        """
        return (X - X.expanding().mean()) / (X.expanding().std())


class Accumulate(TransformerStep):
    """
    Compute the accumulated sum of each feature.

    This is useful to compute the count of binary features.

    Parameters:
        normalize: Weather to apply the normalization or not, by default False
    """ 

    def __init__(self, *, normalize: bool = False,  name: Optional[str] = None):
        super().__init__(name=name)
        self.normalize = normalize

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life computing the cumulated sum

        Parameters:
            X: The Input life
            
        Returns:
            Return a new DataFrame with the same index as the input with the cumulated sum of the features
        """
        X1 = X.cumsum()
        if self.normalize:
            return X1 / X1.abs().apply(np.sqrt)
        else:
            return X1


class Diff(TransformerStep):
    """Compute the 1 step difference of each feature."""

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life computing the 1 step difference

        Parameters:
            X: The Input life

        Returns:
            Return a new DataFrame with the same index as the input with the 1 step difference of the features
        """
        return X.diff()


class StringConcatenate(TransformerStep):
    """
    Return a new DataFrame with a single column with the concatenation of the values of each row. The method works only on strings and the values are separated by -"""

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life computing the 1 step difference

        Parameters:
            X: The Input life
            
        Returns:
            Return a new DataFrame with the same index as the input with the difference of the features
        """
        new_X = pd.DataFrame(index=X.index)
        new_X["concatenation"] = X.agg("-".join, axis=1)
        return new_X


class Apply(TransformerStep):
    """Apply the input function element-wise"""

    def __init__(self, *, fun, name: Optional[str] = None):
        super().__init__(name=name)
        self.fun = fun

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life applying the input function to each element

        Parameters
            X: The Input life
            
        Returns:
            Return a new DataFrame with results of the function application to each element.
        """
        return X.apply(self.fun)


class Clip(TransformerStep):
    """
    Clip values onto a predefined range

    Parameters:
        lower: The lower value
        upper: The Upper value
    """

    def __init__(self, *, lower: float, upper: float, **kwargs):
        super().__init__(**kwargs)
        self.lower = lower
        self.upper = upper

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life clipping the values onto a predefined range

        Parameters:
            X: The Input life
        
        Returns:
            Return a new DataFrame with the same index as the input with the clipped values
        """
        return X.clip(self.lower, self.upper)


class SubstractLinebase(TransformerStep):
    """Subtract the values in the first row from all the rows in the input life"""

    def __init__(self, *args):
        super().__init__(*args)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life

        Parameters:
            X: The Input life

        Returns:
            Return a new DataFrame with the same index as the input with the subtraction of the first row
        """
        return X - X.iloc[0, :]


class Peaks(TransformerStep):
    """Find Peaks in the input life. Return a new DataFrame with the same index as the input with 1 in the position of the peaks and 0 otherwise"""

    distance: float

    def __init__(self, *, distance:float, name : Optional[str] = None):
        super().__init__(name=name)
        self.distance = distance

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Detect the peaks in the input life

        Parameters:
            X: The Input life
        
        Returns:
            Return a new DataFrame with the same index as the input with the peaks marked as 1 and 0 otherwise. 
        """
        new_X = pd.DataFrame(np.zeros(X.shape), index=X.index, columns=X.columns)
        for i, c in enumerate(X.columns):
            peaks_positions, _ = find_peaks(X[c].values, distance=self.distance)
            new_X.iloc[peaks_positions, i] = 1

        return new_X



"""
Module Name : curlyBrace

Author : 高斯羽 博士 (Dr. GAO, Siyu)

Version : 1.0.2

Last Modified : 2019-04-22

This module is basically an Python implementation of the function written Pål Næverlid Sævik
for MATLAB (link in Reference).

The function "curlyBrace" allows you to plot an optionally annotated curly bracket between 
two points when using matplotlib.

The usual settings for line and fonts in matplotlib also apply.

The function takes the axes scales into account automatically. But when the axes aspect is 
set to "equal", the auto switch should be turned off.

Change Log
----------------------
* **Notable changes:**
    + Version : 1.0.2
        - Added considerations for different scaled axes and log scale
    + Version : 1.0.1
        - First version.

Reference
----------------------
https://uk.mathworks.com/matlabcentral/fileexchange/38716-curly-brace-annotation

List of functions
----------------------

* getAxSize_
* curlyBrace_

"""

import matplotlib.pyplot as plt
import numpy as np


def getAxSize(fig, ax):
    """
    .. _getAxSize :

    Get the axes size in pixels.

    Parameters
    ----------
    fig : matplotlib figure object
        The of the target axes.

    ax : matplotlib axes object
        The target axes.

    Returns
    -------
    ax_width : float
        The axes width in pixels.

    ax_height : float
        The axes height in pixels.

    Reference
    -----------
    https://stackoverflow.com/questions/19306510/determine-matplotlib-axis-size-in-pixels
    """

    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    ax_width, ax_height = bbox.width, bbox.height
    ax_width *= fig.dpi
    ax_height *= fig.dpi

    return ax_width, ax_height


def curlyBrace(
    fig,
    ax,
    p1,
    p2,
    k_r=0.1,
    bool_auto=True,
    str_text="",
    int_line_num=2,
    fontdict={},
    **kwargs
):
    # def curlyBrace(fig, ax, p1, p2, k_r=0.1, bool_auto=True, str_text='', int_line_num=2, fontdict={}, **kwargs):
    """
    .. _curlyBrace :

    Plot an optionally annotated curly bracket on the given axes of the given figure.

    Note that the brackets are anti-clockwise by default. To reverse the text position, swap
    "p1" and "p2".

    Note that, when the axes aspect is not set to "equal", the axes coordinates need to be
    transformed to screen coordinates, otherwise the arcs may not be seeable.

    Parameters
    ----------
    fig : matplotlib figure object
        The of the target axes.

    ax : matplotlib axes object
        The target axes.

    p1 : two element numeric list
        The coordinates of the starting point.

    p2 : two element numeric list
        The coordinates of the end point.

    k_r : float
        This is the gain controlling how "curvy" and "pointy" (height) the bracket is.

        Note that, if this gain is too big, the bracket would be very strange.

    bool_auto : boolean
        This is a switch controlling wether to use the auto calculation of axes
        scales.

        When the two axes do not have the same aspects, i.e., not "equal" scales,
        this should be turned on, i.e., True.

        When "equal" aspect is used, this should be turned off, i.e., False.

        If you do not set this to False when setting the axes aspect to "equal",
        the bracket will be in funny shape.

        Default = True

    str_text : string
        The annotation text of the bracket. It would displayed at the mid point
        of bracket with the same rotation as the bracket.

        By default, it follows the anti-clockwise convention. To flip it, swap
        the end point and the starting point.

        The appearance of this string can be set by using "fontdict", which follows
        the same syntax as the normal matplotlib syntax for font dictionary.

        Default = empty string (no annotation)

    int_line_num : int
        This argument determines how many lines the string annotation is from the summit
        of the bracket.

        The distance would be affected by the font size, since it basically just a number of
        lines appended to the given string.

        Default = 2

    fontdict : dictionary
        This is font dictionary setting the string annotation. It is the same as normal
        matplotlib font dictionary.

        Default = empty dict

    **kwargs : matplotlib line setting arguments
        This allows the user to set the line arguments using named arguments that are
        the same as in matplotlib.

    Returns
    -------
    theta : float
        The bracket angle in radians.

    summit : list
        The positions of the bracket summit.

    arc1 : list of lists
        arc1 positions.

    arc2 : list of lists
        arc2 positions.

    arc3 : list of lists
        arc3 positions.

    arc4 : list of lists
        arc4 positions.

    Reference
    ----------
    https://uk.mathworks.com/matlabcentral/fileexchange/38716-curly-brace-annotation
    """

    pt1 = [None, None]
    pt2 = [None, None]

    ax_width, ax_height = getAxSize(fig, ax)

    ax_xlim = list(ax.get_xlim())
    ax_ylim = list(ax.get_ylim())

    # log scale consideration
    if "log" in ax.get_xaxis().get_scale():

        if p1[0] > 0.0:

            pt1[0] = np.log(p1[0])

        elif p1[0] < 0.0:

            pt1[0] = -np.log(abs(p1[0]))

        else:

            pt1[0] = 0.0

        if p2[0] > 0.0:

            pt2[0] = np.log(p2[0])

        elif p2[0] < 0.0:

            pt2[0] = -np.log(abs(p2[0]))

        else:

            pt2[0] = 0

        for i in range(0, len(ax_xlim)):

            if ax_xlim[i] > 0.0:

                ax_xlim[i] = np.log(ax_xlim[i])

            elif ax_xlim[i] < 0.0:

                ax_xlim[i] = -np.log(abs(ax_xlim[i]))

            else:

                ax_xlim[i] = 0.0

    else:

        pt1[0] = p1[0]
        pt2[0] = p2[0]

    if "log" in ax.get_yaxis().get_scale():

        if p1[1] > 0.0:

            pt1[1] = np.log(p1[1])

        elif p1[1] < 0.0:

            pt1[1] = -np.log(abs(p1[1]))

        else:

            pt1[1] = 0.0

        if p2[1] > 0.0:

            pt2[1] = np.log(p2[1])

        elif p2[1] < 0.0:

            pt2[1] = -np.log(abs(p2[1]))

        else:

            pt2[1] = 0.0

        for i in range(0, len(ax_ylim)):

            if ax_ylim[i] > 0.0:

                ax_ylim[i] = np.log(ax_ylim[i])

            elif ax_ylim[i] < 0.0:

                ax_ylim[i] = -np.log(abs(ax_ylim[i]))

            else:

                ax_ylim[i] = 0.0

    else:

        pt1[1] = p1[1]
        pt2[1] = p2[1]

    # get the ratio of pixels/length
    xscale = ax_width / abs(ax_xlim[1] - ax_xlim[0])
    yscale = ax_height / abs(ax_ylim[1] - ax_ylim[0])

    # this is to deal with 'equal' axes aspects
    if bool_auto:

        pass

    else:

        xscale = 1.0
        yscale = 1.0

    # convert length to pixels,
    # need to minus the lower limit to move the points back to the origin. Then add the limits back on end.
    pt1[0] = (pt1[0] - ax_xlim[0]) * xscale
    pt1[1] = (pt1[1] - ax_ylim[0]) * yscale
    pt2[0] = (pt2[0] - ax_xlim[0]) * xscale
    pt2[1] = (pt2[1] - ax_ylim[0]) * yscale

    # calculate the angle
    theta = np.arctan2(pt2[1] - pt1[1], pt2[0] - pt1[0])

    # calculate the radius of the arcs
    r = np.hypot(pt2[0] - pt1[0], pt2[1] - pt1[1]) * k_r

    # arc1 centre
    x11 = pt1[0] + r * np.cos(theta)
    y11 = pt1[1] + r * np.sin(theta)

    # arc2 centre
    x22 = (pt2[0] + pt1[0]) / 2.0 - 2.0 * r * np.sin(theta) - r * np.cos(theta)
    y22 = (pt2[1] + pt1[1]) / 2.0 + 2.0 * r * np.cos(theta) - r * np.sin(theta)

    # arc3 centre
    x33 = (pt2[0] + pt1[0]) / 2.0 - 2.0 * r * np.sin(theta) + r * np.cos(theta)
    y33 = (pt2[1] + pt1[1]) / 2.0 + 2.0 * r * np.cos(theta) + r * np.sin(theta)

    # arc4 centre
    x44 = pt2[0] - r * np.cos(theta)
    y44 = pt2[1] - r * np.sin(theta)

    # prepare the rotated
    q = np.linspace(theta, theta + np.pi / 2.0, 50)

    # reverse q
    # t = np.flip(q) # this command is not supported by lower version of numpy
    t = q[::-1]

    # arc coordinates
    arc1x = r * np.cos(t + np.pi / 2.0) + x11
    arc1y = r * np.sin(t + np.pi / 2.0) + y11

    arc2x = r * np.cos(q - np.pi / 2.0) + x22
    arc2y = r * np.sin(q - np.pi / 2.0) + y22

    arc3x = r * np.cos(q + np.pi) + x33
    arc3y = r * np.sin(q + np.pi) + y33

    arc4x = r * np.cos(t) + x44
    arc4y = r * np.sin(t) + y44

    # convert back to the axis coordinates
    arc1x = arc1x / xscale + ax_xlim[0]
    arc2x = arc2x / xscale + ax_xlim[0]
    arc3x = arc3x / xscale + ax_xlim[0]
    arc4x = arc4x / xscale + ax_xlim[0]

    arc1y = arc1y / yscale + ax_ylim[0]
    arc2y = arc2y / yscale + ax_ylim[0]
    arc3y = arc3y / yscale + ax_ylim[0]
    arc4y = arc4y / yscale + ax_ylim[0]

    # log scale consideration
    if "log" in ax.get_xaxis().get_scale():

        for i in range(0, len(arc1x)):

            if arc1x[i] > 0.0:

                arc1x[i] = np.exp(arc1x[i])

            elif arc1x[i] < 0.0:

                arc1x[i] = -np.exp(abs(arc1x[i]))

            else:

                arc1x[i] = 0.0

        for i in range(0, len(arc2x)):

            if arc2x[i] > 0.0:

                arc2x[i] = np.exp(arc2x[i])

            elif arc2x[i] < 0.0:

                arc2x[i] = -np.exp(abs(arc2x[i]))

            else:

                arc2x[i] = 0.0

        for i in range(0, len(arc3x)):

            if arc3x[i] > 0.0:

                arc3x[i] = np.exp(arc3x[i])

            elif arc3x[i] < 0.0:

                arc3x[i] = -np.exp(abs(arc3x[i]))

            else:

                arc3x[i] = 0.0

        for i in range(0, len(arc4x)):

            if arc4x[i] > 0.0:

                arc4x[i] = np.exp(arc4x[i])

            elif arc4x[i] < 0.0:

                arc4x[i] = -np.exp(abs(arc4x[i]))

            else:

                arc4x[i] = 0.0

    else:

        pass

    if "log" in ax.get_yaxis().get_scale():

        for i in range(0, len(arc1y)):

            if arc1y[i] > 0.0:

                arc1y[i] = np.exp(arc1y[i])

            elif arc1y[i] < 0.0:

                arc1y[i] = -np.exp(abs(arc1y[i]))

            else:

                arc1y[i] = 0.0

        for i in range(0, len(arc2y)):

            if arc2y[i] > 0.0:

                arc2y[i] = np.exp(arc2y[i])

            elif arc2y[i] < 0.0:

                arc2y[i] = -np.exp(abs(arc2y[i]))

            else:

                arc2y[i] = 0.0

        for i in range(0, len(arc3y)):

            if arc3y[i] > 0.0:

                arc3y[i] = np.exp(arc3y[i])

            elif arc3y[i] < 0.0:

                arc3y[i] = -np.exp(abs(arc3y[i]))

            else:

                arc3y[i] = 0.0

        for i in range(0, len(arc4y)):

            if arc4y[i] > 0.0:

                arc4y[i] = np.exp(arc4y[i])

            elif arc4y[i] < 0.0:

                arc4y[i] = -np.exp(abs(arc4y[i]))

            else:

                arc4y[i] = 0.0

    else:

        pass

    # plot arcs
    ax.plot(arc1x, arc1y, clip_on=False, **kwargs)
    ax.plot(arc2x, arc2y, clip_on=False, **kwargs)
    ax.plot(arc3x, arc3y, clip_on=False, **kwargs)
    ax.plot(arc4x, arc4y, clip_on=False, **kwargs)

    # plot lines
    ax.plot([arc1x[-1], arc2x[1]], [arc1y[-1], arc2y[1]], clip_on=False, **kwargs)
    ax.plot([arc3x[-1], arc4x[1]], [arc3y[-1], arc4y[1]], clip_on=False, **kwargs)

    summit = [arc2x[-1], arc2y[-1]]

    if str_text:

        int_line_num = int(int_line_num)

        str_temp = "\n" * int_line_num

        # convert radians to degree and within 0 to 360
        ang = np.degrees(theta) % 360.0

        if (ang >= 0.0) and (ang <= 90.0):

            rotation = ang

            str_text = str_text + str_temp

        if (ang > 90.0) and (ang < 270.0):

            rotation = ang + 180.0

            str_text = str_temp + str_text

        elif (ang >= 270.0) and (ang <= 360.0):

            rotation = ang

            str_text = str_text + str_temp

        else:

            rotation = ang

        ax.axes.text(
            arc2x[-1],
            arc2y[-1],
            str_text,
            ha="center",
            va="center",
            rotation=rotation,
            fontdict=fontdict,
            clip_on=False,
        )

    else:

        pass

    arc1 = [arc1x, arc1y]
    arc2 = [arc2x, arc2y]
    arc3 = [arc3x, arc3y]
    arc4 = [arc4x, arc4y]

    return theta, summit, arc1, arc2, arc3, arc4



from pathlib import Path

import numpy as np
import pandas as pd
import tensorflow as tf
from numpy.random import seed
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import make_pipeline
from tensorflow.keras import Input, Model
from tensorflow.keras.layers import Dense, Flatten
from xgboost import XGBRegressor

from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.iterators.iterators import WindowedDatasetIterator
from ceruleo.iterators.sample_weight import RULInverseWeighted
from ceruleo.iterators.shufflers import AllShuffled
from ceruleo.iterators.utils import true_values
from ceruleo.models.baseline import BaselineModel, FixedValueBaselineModel
from ceruleo.models.keras.callbacks import PredictionCallback
from ceruleo.models.keras.catalog.CNLSTM import CNLSTM
from ceruleo.models.keras.catalog.InceptionTime import InceptionTime
from ceruleo.models.keras.catalog.MSWRLRCN import MSWRLRCN
from ceruleo.models.keras.catalog.MultiScaleConvolutional import (
    MultiScaleConvolutionalModel,
)
from ceruleo.models.keras.catalog.XCM import XCM, explain
from ceruleo.models.keras.catalog.XiangQiangJianQiao import XiangQiangJianQiaoModel
from ceruleo.models.keras.dataset import tf_regression_dataset
from ceruleo.models.keras.losses import (
    AsymmetricLossPM,
    asymmetric_loss_pm,
    relative_mae,
    relative_mse,
    root_mean_squared_error,
)
from ceruleo.models.sklearn import (
    CeruleoRegressor,
    EstimatorWrapper,
    TimeSeriesWindowTransformer,
    predict,
    train_model,
)
from ceruleo.transformation import Transformer
from ceruleo.transformation.features.scalers import MinMaxScaler
from ceruleo.transformation.features.selection import ByNameFeatureSelector

seed(1)


tf.random.set_seed(2)


class SimpleDataset(AbstractPDMDataset):
    def __init__(self):

        self.lives = [
            pd.DataFrame(
                {"feature1": np.array(range(0, 100)), "RUL": np.array(range(0, 100))}
            )
        ]

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


class MockDataset(AbstractPDMDataset):
    def __init__(self, nlives: int):

        self.lives = [
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, 5 * 100, 50),
                    "feature2": np.linspace(-25, 5 * 500, 50),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
            for i in range(nlives - 1)
        ]

        self.lives.append(
            pd.DataFrame(
                {
                    "feature1": np.linspace(0, 5 * 100, 50),
                    "feature2": np.linspace(-25, 5 * 500, 50),
                    "feature3": np.linspace(-25, 5 * 500, 50),
                    "RUL": np.linspace(100, 0, 50),
                }
            )
        )

    def get_time_series(self, i: int):
        return self.lives[i]

    @property
    def rul_column(self):
        return "RUL"

    @property
    def n_time_series(self):
        return len(self.lives)


def _test_model_basic(model, ds_iterator, loss="mae"):
    model.compile(loss=loss, optimizer=tf.keras.optimizers.SGD(0.0001))
    model.fit(tf_regression_dataset(ds_iterator).batch(15), verbose=False)
    y_pred = model.predict(tf_regression_dataset(ds_iterator).batch(15)).ravel()
    assert isinstance(y_pred, np.ndarray)


class TestModels:
    def test_models(self):
        features = ["feature1", "feature2"]
        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(x, y)
        batch_size = 15
        window_size = 5
        ds = MockDataset(5)
        transformer.fit(ds)
        iterator = WindowedDatasetIterator(
            ds.map(transformer),
            window_size,
            step=1,
            horizon=1,
        )

        b1 = tf_regression_dataset(iterator).batch(15)
        assert b1.take(1)

    def test_sklearn(self):
        features = ["feature1", "feature2"]
        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(x, y)
        ds = MockDataset(5)

        window_transformer = TimeSeriesWindowTransformer(transformer, window_size=5)
        pipe = make_pipeline(window_transformer, EstimatorWrapper(LinearRegression()))
        pipe.fit(ds)

        y_pred = pipe.predict(ds)
        y_true = window_transformer.true_values(ds)

        mse = np.sum((y_pred - y_true) ** 2)
        assert mse < 0.01

        regressor = CeruleoRegressor(
            TimeSeriesWindowTransformer(transformer, window_size=15, step=1),
            LinearRegression(),
        )

        regressor.fit(ds)
        y_pred = regressor.predict(ds)
        y_true = regressor.ts_window_transformer.true_values(ds)
        mse = np.sum((y_pred - y_true) ** 2)
        assert mse < 0.01

    def test_keras(self):

        features = ["feature1", "feature2"]
        pipe = ByNameFeatureSelector(features=features)
        pipe = MinMaxScaler(range=(-1, 1))(pipe)
        rul_pipe = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(pipe, rul_pipe)
        ds = MockDataset(5)
        transformer.fit(ds)
        train_dataset = ds[range(0, 4)]
        val_dataset = ds[range(4, 5)]
        window_size = 6
        train_iterator = WindowedDatasetIterator(
            train_dataset.map(transformer),
            window_size=window_size,
            step=1,
            shuffler=AllShuffled(),
        )

        val_iterator = WindowedDatasetIterator(
            val_dataset.map(transformer), window_size=window_size, step=1
        )

        input = Input(shape=train_iterator.shape)
        x = input
        x = Flatten()(x)
        x = Dense(8, activation="relu")(x)
        x = Dense(5, activation="relu")(x)
        x = Dense(1)(x)
        model = Model(inputs=[input], outputs=[x])
        model.compile(loss="mae", optimizer=tf.keras.optimizers.SGD(0.01))
        y_true = true_values(val_iterator)
        y_pred_before_fit = model.predict(tf_regression_dataset(val_iterator).batch(64))
        mae_before_fit = np.mean(np.abs(y_pred_before_fit.ravel() - y_true.ravel()))
        model.fit(
            tf_regression_dataset(train_iterator).batch(4),
            validation_data=tf_regression_dataset(val_iterator).batch(64),
            epochs=15,
        )
        y_pred = model.predict(tf_regression_dataset(val_iterator).batch(64))

        mae = np.mean(np.abs(y_pred.ravel() - y_true.ravel()))

        assert mae < mae_before_fit

        train_iterator = WindowedDatasetIterator(
            train_dataset.map(transformer),
            window_size=window_size,
            step=1,
            shuffler=AllShuffled(),
            sample_weight=RULInverseWeighted(),
        )

        val_iterator = WindowedDatasetIterator(
            val_dataset.map(transformer), window_size=window_size, step=1
        )
        model.compile(loss="mae", optimizer=tf.keras.optimizers.SGD(0.01))
        y_true = true_values(val_iterator)
        y_pred_before_fit = model.predict(tf_regression_dataset(val_iterator).batch(64))
        mae_before_fit = np.mean(np.abs(y_pred_before_fit.ravel() - y_true.ravel()))
        model.fit(
            tf_regression_dataset(train_iterator).batch(4),
            validation_data=tf_regression_dataset(val_iterator).batch(64),
            epochs=15,
        )
        y_pred = model.predict(tf_regression_dataset(val_iterator).batch(64))

        mae = np.mean(np.abs(y_pred.ravel() - y_true.ravel()))

        assert mae < mae_before_fit
        
    def test_xgboost(self):
        features = ["feature1", "feature2"]

        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(x, y)

        ds = MockDataset(5)
        transformer.fit(ds)
        transformed_ds = ds.map(transformer)
        ds_iterator = WindowedDatasetIterator(
            transformed_ds, window_size=5, step=2, shuffler=AllShuffled()
        )
        model = XGBRegressor(n_estimators=500)
        train_model(model, ds_iterator)

        y_pred = predict(model, ds_iterator)
        y_true = true_values(ds_iterator)

        assert np.sum(y_pred - y_true) < 0.001

    def test_catalog(self):
        features = ["feature1", "feature2"]

        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(x, y)

        ds = MockDataset(5)
        transformer.fit(ds)
        transformed_ds = ds.map(transformer)
        ds_iterator = WindowedDatasetIterator(
            transformed_ds, window_size=5, step=2, shuffler=AllShuffled()
        )

        model = CNLSTM(
            ds_iterator.shape,
            n_conv_layers=2,
            initial_convolutional_size=5,
            layers_recurrent=[5, 5],
            hidden_size=(15, 5),
            dropout=0.3,
        )
        _test_model_basic(model, ds_iterator)

        model = InceptionTime(
            ds_iterator.shape,
            nb_filters=3,
        )
        _test_model_basic(model, ds_iterator)

        model = MSWRLRCN(ds_iterator.shape)
        _test_model_basic(model, ds_iterator)

        model = MultiScaleConvolutionalModel(
            ds_iterator.shape, n_msblocks=1, scales=[2, 3], n_hidden=5
        )
        _test_model_basic(model, ds_iterator)

        model = XiangQiangJianQiaoModel(ds_iterator.shape)
        _test_model_basic(model, ds_iterator)

        model, model_extras = XCM(ds_iterator.shape)
        _test_model_basic(model, ds_iterator)
        X, y, sw = next(iter(ds_iterator))
        (mmap, v) = explain(model_extras, X)
        print(type(mmap))
        assert isinstance(mmap, np.ndarray)

        #model, model_extras = MVCNN(ds_iterator.shape)
        #_test_model_basic(model, ds_iterator)
        #X, y, sw = next(iter(ds_iterator))
        #(mmap, v) = explain(model_extras, X)
        #print(type(mmap))
        #assert isinstance(mmap, np.ndarray)

    def test_baseline(self):
        ds = MockDataset(5)
        features = ["feature1", "feature2"]

        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(x, y)

        transformer.fit(ds)
        transformed_ds = ds.map(transformer)

        model_mean = BaselineModel(mode="mean")
        model_mean.fit(ds)
        y_pred = model_mean.predict(ds)
        assert isinstance(y_pred, np.ndarray)

        model_median = BaselineModel(mode="median")
        model_median.fit(ds)
        y_pred = model_mean.predict(ds)
        assert isinstance(y_pred, np.ndarray)

        model_fixed = FixedValueBaselineModel(value=100)
        model_fixed.fit(ds)
        y_pred = model_mean.predict(ds)
        assert isinstance(y_pred, np.ndarray)

        model_mean = BaselineModel(mode="mean")
        model_mean.fit(transformed_ds)
        y_pred = model_mean.predict(transformed_ds)
        assert isinstance(y_pred, np.ndarray)

        model_median = BaselineModel(mode="median")
        model_median.fit(transformed_ds)
        y_pred = model_mean.predict(transformed_ds)
        assert isinstance(y_pred, np.ndarray)

        model_fixed = FixedValueBaselineModel(value=100)
        model_fixed.fit(transformed_ds)
        y_pred = model_mean.predict(transformed_ds)
        assert isinstance(y_pred, np.ndarray)

    def test_callbacks(self):
        features = ["feature1", "feature2"]

        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(x, y)

        ds = MockDataset(5)
        transformer.fit(ds)
        transformed_ds = ds.map(transformer)
        ds_iterator = WindowedDatasetIterator(
            transformed_ds, window_size=5, step=2, shuffler=AllShuffled()
        )
        model = MultiScaleConvolutionalModel(
            ds_iterator.shape, n_msblocks=1, scales=[2, 3], n_hidden=5
        )
        model.compile(loss="mae", optimizer=tf.keras.optimizers.SGD(0.0001))

        tf_dataset = tf_regression_dataset(ds_iterator).batch(15)
        output_path = Path(".").resolve() / "output.png"
        model.fit(
            tf_dataset,
            callbacks=[PredictionCallback(output_path, tf_dataset)],
            verbose=False,
        )

        assert output_path.is_file()

    def test_losses(self):
        features = ["feature1", "feature2"]

        x = ByNameFeatureSelector(features=features)
        x = MinMaxScaler(range=(-1, 1))(x)

        y = ByNameFeatureSelector(features=["RUL"])
        transformer = Transformer(x, y)

        ds = MockDataset(5)
        transformer.fit(ds)
        transformed_ds = ds.map(transformer)
        ds_iterator = WindowedDatasetIterator(
            transformed_ds, window_size=5, step=2, shuffler=AllShuffled()
        )
        model = MultiScaleConvolutionalModel(
            ds_iterator.shape, n_msblocks=1, scales=[2, 3], n_hidden=5
        )
        _test_model_basic(
            model,
            ds_iterator,
            loss=AsymmetricLossPM(
                theta_l=1, alpha_l=1, gamma_l=1, theta_r=1, alpha_r=1, gamma_r=1
            ),
        )

        _test_model_basic(model, ds_iterator, loss=root_mean_squared_error)

        _test_model_basic(model, ds_iterator, loss=relative_mae(C=0.5))

        _test_model_basic(model, ds_iterator, loss=relative_mse(C=0.5))

        assert isinstance(
            root_mean_squared_error(tf.random.uniform((50,)), tf.random.uniform((50,))).numpy(), np.float32
        )
        assert isinstance(
            asymmetric_loss_pm(
                tf.random.uniform((50,)),
                tf.random.uniform((50,)),
                theta_l=1,
                alpha_l=1,
                gamma_l=1,
                theta_r=1,
                alpha_r=1,
                gamma_r=1,
            ).numpy(),
            np.float32,
        )


  
